"""
This module implements the folding abstraction.
"""

import os
import glob
import shutil

import numpy as np
from sklearn.model_selection import RepeatedStratifiedKFold

from ..base import dump_dict

__all__= ['Folding']

class Folding():
    """
    Cache-able folding of dataset for cross-validation
    """

    def __init__(self,
                dataset,
                *,
                serialization='json',
                cache_path=None,
                validator_params=None,
                reset=False):
        """
        Constructor of Folding object

        Args:
            dataset (dict): dataset dictionary with keys 'data', 'target'
                            and 'DESCR'
            validator (obj): cross-validator object
            serialization (str): 'json'/'pickle'
            cache_path (str): path to cache directory
            validator_params (str): validator parameters
        """
        self.dataset = dataset
        if validator_params is None:
            self.validator_params = {'n_repeats': 2,
                                        'n_splits': 5,
                                        'random_state': 5}
        else:
            self.validator_params = validator_params
        self.serialization = serialization
        self.cache_path = cache_path
        self.reset = reset

        labels, counts = np.unique(self.dataset['target'],
                                    return_counts=True)

        min_label = int(np.max(labels[counts == np.min(counts)]))
        maj_label = int(labels[labels != min_label][0])

        n_min = int(counts[min_label])
        n_maj = int(counts[maj_label])
        imb_ratio = float(counts[maj_label]/counts[min_label])

        self.properties = {'db_n': len(dataset['data']),
                            'db_n_attr': len(dataset['data'][0]),
                            'imbalance_ratio': imb_ratio,
                            'name': dataset['name'],
                            'label_stats': {'min_label': n_min,
                                            'maj_label': n_maj}}

        if self.cache_path is not None:
            self.folding_dir_path = os.path.join(self.cache_path, dataset['name'])

            # creating the folding directory
            os.makedirs(self.folding_dir_path, exist_ok=True)

            if reset:
                self._do_reset()

    def _do_reset(self):
        """
        Remove all foldings
        """
        shutil.rmtree(self.folding_dir_path, ignore_errors=True)
        os.makedirs(self.folding_dir_path, exist_ok=True)

    def folding_files(self):
        """
        Get the list of filenames

        Returns:
            list: the list of fold filenames
        """
        if self.serialization == 'json':
            path = os.path.join(self.folding_dir_path, 'fold_*.json')
        elif self.serialization == 'pickle':
            path = os.path.join(self.folding_dir_path, 'fold_*.pickle')

        if len(glob.glob(path)) == 0:
            self.cache_foldings()

        return sorted(glob.glob(path))

    def fold(self):
        """
        Does the folding or reads it from file if already available

        Returns:
            list(tuple): list of tuples of X_train, y_train, X_test, y_test
                            objects
        """
        validator = RepeatedStratifiedKFold(**self.validator_params)

        data = self.dataset['data']
        target = self.dataset['target']

        properties = self.properties
        n_splits = self.validator_params['n_splits']
        properties['n_repeats'] = self.validator_params['n_repeats']
        properties['n_splits'] = n_splits

        for idx, (train, test) in enumerate(validator.split(data,
                                                            target,
                                                            target)):
            properties['split_idx'] = idx
            properties['repeat_idx'] = idx // n_splits
            properties['fold_idx'] = idx - (idx // n_splits)*n_splits

            folding = {'X_train': data[train],
                        'y_train': target[train],
                        'X_test': data[test],
                        'y_test': target[test],
                        'fold_descriptor': properties}

            yield folding

    def cache_foldings(self):
        """
        Cache the foldings
        """
        for folding in self.fold():
            repeat_idx = folding['fold_descriptor']['repeat_idx']
            fold_idx = folding['fold_descriptor']['fold_idx']
            filename = f"fold_{repeat_idx:04}_{fold_idx:04}"\
                        f".{self.serialization}"
            full_name = os.path.join(self.folding_dir_path, filename)

            dump_dict(folding, full_name, self.serialization,
                        ['X_train', 'X_test', 'y_train', 'y_test'])

    def get_params(self, deep=False):
        """
        Return the parameters of the object

        Args:
            deep (bool): deep parameters

        Returns:
            dict: the parameters of the object
        """
        _ = deep # pylint fix

        return {'dataset': self.dataset,
                'serialization': self.serialization,
                'cache_path': self.cache_path,
                'validator_params': self.validator_params,
                'reset': self.reset}

    def descriptor(self):
        """
        Returns the descriptor of the object

        Returns:
            str: the descriptor of the object
        """
        return str(self.get_params())



"""
This module implements the Gaussian_SMOTE method.
"""

import numpy as np

from sklearn.preprocessing import StandardScaler

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['Gaussian_SMOTE']

class Gaussian_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{gaussian_smote,
                      title={Gaussian-Based SMOTE Algorithm for Solving Skewed
                                Class Distributions},
                      author={Hansoo Lee and Jonggeun Kim and Sungshin Kim},
                      journal={Int. J. Fuzzy Logic and Intelligent Systems},
                      year={2017},
                      volume={17},
                      pages={229-234}
                    }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 sigma=1.0,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            sigma (float): variance
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': {'sigma': sigma}}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater(sigma, "sigma", 0.0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.sigma = sigma
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'sigma': [0.5, 1.0, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # standardization applied to make sigma compatible with the data
        scaler = StandardScaler()
        X_ss = scaler.fit_transform(X) # pylint: disable=invalid-name

        # fitting nearest neighbors model to find the minority neighbors of
        # minority samples
        X_min = X_ss[y == self.min_label]
        n_neighbors = min([len(X_min), self.n_neighbors + 1])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        samples = self.sample_simplex(X=X_min,
                                        indices=ind,
                                        n_to_sample=n_to_sample)


        return (np.vstack([X, scaler.inverse_transform(np.vstack(samples))]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'sigma': self.sigma,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Tests the multiclass oversampling functionalities.
"""

import logging

import numpy as np
from sklearn import datasets
import pytest

import smote_variants as sv

logger = logging.getLogger("smote_variants")
logger.setLevel(logging.ERROR)

def test_multiclass():
    """
    Test the multiclass oversampling.
    """
    dataset = datasets.load_wine()

    oversampler = sv.MulticlassOversampling('SMOTE')

    X_samp, y_samp = oversampler.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

    _, counts = np.unique(y_samp, return_counts=True)

    assert np.all(counts == counts[0])

    oversampler = sv.MulticlassOversampling('SMOTE',
                                            strategy='equalize_1_vs_many')

    X_samp, y_samp = oversampler.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

def test_non_compatible():
    """
    Test the use of incompatible oversampler.
    """
    with pytest.raises(ValueError):
        sv.MulticlassOversampling('SUNDO')

def test_nonsense():
    """
    Test the use of meaningless strategy.
    """
    with pytest.raises(ValueError):
        sv.MulticlassOversampling('SMOTE', strategy='nonsense')

def test_get_params():
    """
    Test the get_params function.
    """
    params = sv.MulticlassOversampling('SMOTE').get_params()
    assert len(params) > 0



"""
Testing the SMOTE_PSOBAT.
"""
import numpy as np

from smote_variants import SMOTE_PSOBAT

def test_specific():
    """
    Oversampler specific testing
    """
    obj = SMOTE_PSOBAT(method='pso')

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9]])
    y = np.array([0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0

    assert len(obj.parameter_combinations(raw=True)) > 0



"""
This module implements the CNN TomekLinks technique.
"""

from ._noisefilter import NoiseFilter
from ._tomeklinkremoval import TomekLinkRemoval
from ._condensednn import CondensedNearestNeighbors

from .._logger import logger
_logger= logger

__all__= ['CNNTomekLinks']

class CNNTomekLinks(NoiseFilter):
    """
    References:
        * BibTex::

            @article{smoteNoise0,
                     author = {Batista, Gustavo E. A. P. A. and Prati,
                                Ronaldo C. and Monard, Maria Carolina},
                     title = {A Study of the Behavior of Several Methods
                                for Balancing Machine Learning Training Data},
                     journal = {SIGKDD Explor. Newsl.},
                     issue_date = {June 2004},
                     volume = {6},
                     number = {1},
                     month = jun,
                     year = {2004},
                     issn = {1931-0145},
                     pages = {20--29},
                     numpages = {10},
                     url = {http://doi.acm.org/10.1145/1007730.1007735},
                     doi = {10.1145/1007730.1007735},
                     acmid = {1007735},
                     publisher = {ACM},
                     address = {New York, NY, USA}
                    }
    """

    def __init__(self, n_jobs=1):
        """
        Constructor of the noise removal object

        Args:
            n_jobs (int): number of parallel jobs
        """
        super().__init__()

        self.check_n_jobs(n_jobs, 'n_jobs')

        self.n_jobs = n_jobs

    def get_params(self, deep=False):
        return {'n_jobs': self.n_jobs,
                **NoiseFilter.get_params(self, deep)}

    def remove_noise(self, X, y):
        """
        Removes noise

        Args:
            X (np.array): features
            y (np.array): target labels

        Returns:
            np.array, np.array: cleaned features and target labels
        """
        _logger.info("%s: Running noise removal.", self.__class__.__name__)
        self.class_label_statistics(y)

        cnn = CondensedNearestNeighbors(n_jobs=self.n_jobs)
        X_cnn, y_cnn = cnn.remove_noise(X, y) # pylint: disable=invalid-name
        tomek = TomekLinkRemoval(n_jobs=self.n_jobs)

        return tomek.remove_noise(X_cnn, y_cnn)



"""
This module implements the MSYN method.
"""

import numpy as np

from ..base import (NearestNeighborsWithMetricTensor,
                            pairwise_distances_mahalanobis, coalesce,
                            coalesce_dict)
from ..base import OverSampling
from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['MSYN']

class MSYN(OverSampling):
    """
    References:
        * BibTex::

            @InProceedings{msyn,
                            author="Fan, Xiannian
                            and Tang, Ke
                            and Weise, Thomas",
                            editor="Huang, Joshua Zhexue
                            and Cao, Longbing
                            and Srivastava, Jaideep",
                            title="Margin-Based Over-Sampling Method for
                                    Learning from Imbalanced Datasets",
                            booktitle="Advances in Knowledge Discovery and
                                        Data Mining",
                            year="2011",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="309--320",
                            abstract="Learning from imbalanced datasets has
                                        drawn more and more attentions from
                                        both theoretical and practical aspects.
                                        Over- sampling is a popular and simple
                                        method for imbalanced learning. In this
                                        paper, we show that there is an
                                        inherently potential risk associated
                                        with the over-sampling algorithms in
                                        terms of the large margin principle.
                                        Then we propose a new synthetic over
                                        sampling method, named Margin-guided
                                        Synthetic Over-sampling (MSYN), to
                                        reduce this risk. The MSYN improves
                                        learning with respect to the data
                                        distributions guided by the
                                        margin-based rule. Empirical study
                                        verities the efficacy of MSYN.",
                            isbn="978-3-642-20847-8"
                            }
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 pressure=1.5,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 proportion=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            pressure (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): number of neighbors in the SMOTE sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(pressure, 'pressure', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.pressure = proportion or pressure
        self.proportion = self.pressure
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, ss_params_default)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [2.5, 2.0, 1.5, 1.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_Delta_P_N(self, *, # pylint: disable=invalid-name
                            nearest_hit_dist,
                            nearest_miss_dist,
                            min_indices,
                            maj_indices,
                            distances,
                            theta_min,
                            theta_maj):
        """
        Determine the Delta_P and Delta_N quantities.

        Args:
            nearest_hit_dist (np.array): the nearest hit distances
            nearest_miss_dist (np.array): the nearest miss distances
            min_indices (np.array): the minority indices
            maj_indices (np.array): the majority indices
            distances (np.array): the distance matrix
            theta_min (np.array): the theta_min array
            theta_maj (np.array): the theta_maj array

        Returns:
            np.array, np.array: the Delta_P and Delta_N arrays
        """

        mask = (nearest_hit_dist[min_indices] < distances[min_indices].T).T
        nearest_hit_dist_min = np.where(np.logical_not(mask).T, distances[min_indices].T,
                                        nearest_hit_dist[min_indices]).T
        nearest_miss_dist_min = nearest_miss_dist[min_indices]
        nearest_hit_dist_maj = nearest_hit_dist[maj_indices]
        mask = (nearest_miss_dist[maj_indices] < distances[maj_indices].T).T
        nearest_miss_dist_maj = np.where(np.logical_not(mask).T, distances[maj_indices].T,
                                            nearest_miss_dist[maj_indices]).T

        theta_x_min = 0.5*(nearest_miss_dist_min - nearest_hit_dist_min.T)
        theta_x_maj = 0.5*(nearest_miss_dist_maj.T - nearest_hit_dist_maj)

        return (np.sum(theta_x_min - theta_min, axis=1),
                np.sum(theta_x_maj - theta_maj, axis=1))

    def determine_nearest_hit_miss_dist(self, X, y, nn_params):
        """
        Determine the nearest hit/miss distances

        Args:
            X (np.array): the feature vectors
            y (np.array): the target labels
            nn_params (dict): the nearest neighbors parameters

        Returns:
            np.array, np.array: the nearest hit/miss distances
        """
        # Compute nearest hit and miss for both classes
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=len(X),
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)
        dist, ind = nnmt.kneighbors(X)

        # computing nearest hit and miss distances, these will be used to
        # compute thetas
        hit_mask = y[ind[:, 1:]] == y[ind[:, 0]][:, None]
        nearest_hit_dist = dist[:, 1:][np.arange(y.shape[0]),
                                        np.argmax(hit_mask, axis=1)]
        nearest_miss_dist = dist[:, 1:][np.arange(y.shape[0]),
                                        np.argmax(np.logical_not(hit_mask),
                                                    axis=1)]

        return nearest_hit_dist, nearest_miss_dist

    def determine_f_3(self, *, X, y, X_new, nn_params):
        """
        Determine the f_3 scores.

        Args:
            X (np.array): all training vectors
            y (np.array): all labels
            X_new (np.array): the new samples
            nn_params (dict): the nearest neighbors parameters

        Returns:
            np.array: the f_3 scores
        """
        min_indices = np.where(y == self.min_label)[0]
        maj_indices = np.where(y == self.maj_label)[0]

        nearest_hit_dist, nearest_miss_dist = \
            self.determine_nearest_hit_miss_dist(X, y, nn_params)

        # computing the thetas without new samples being involved
        theta_A_sub_alpha = 0.5*(nearest_miss_dist - nearest_hit_dist) # pylint: disable=invalid-name
        theta_min = theta_A_sub_alpha[min_indices]
        theta_maj = theta_A_sub_alpha[maj_indices]

        distances = pairwise_distances_mahalanobis(X,
                                                    Y=X_new,
                                                    tensor=nn_params['metric_tensor'])

        Delta_P, Delta_N = self.determine_Delta_P_N(nearest_hit_dist=nearest_hit_dist,  # pylint: disable=invalid-name
                                                    nearest_miss_dist=nearest_miss_dist,
                                                    min_indices=min_indices,
                                                    maj_indices=maj_indices,
                                                    distances=distances,
                                                    theta_min=theta_min,
                                                    theta_maj=theta_maj)

        return -Delta_N/(Delta_P + 0.01)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels
        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.pressure)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        # generating samples
        smote = SMOTE(proportion=self.pressure*3,
                      n_neighbors=self.n_neighbors,
                      nn_params=nn_params,
                      ss_params=self.ss_params,
                      n_jobs=self.n_jobs,
                      random_state=self._random_state_init)

        X_res, y_res = smote.sample(X, y) # pylint: disable=invalid-name
        X_new, _ = X_res[len(X):], y_res[len(X):]

        f_3 = self.determine_f_3(X=X, y=y,
                                X_new=X_new,
                                nn_params=nn_params)

        # determining the elements with the minimum f_3 scores to add
        _, new_ind = zip(*sorted(zip(f_3, np.arange(len(f_3))), key=lambda x: x[0]))
        new_ind = list(new_ind[:n_to_sample])

        return (np.vstack([X, X_new[new_ind]]),
                np.hstack([y, np.repeat(self.min_label, len(new_ind))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'pressure': self.pressure,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self),
                'proportion': self.proportion}



"""
Testing the NoiseFilter class.
"""

from smote_variants.noise_removal import NoiseFilter

def test_noisefilter():
    """
    Tests the noisefilter base class.
    """
    noisef = NoiseFilter()
    noisef.set_params(**{'dummy': 1})

    assert getattr(noisef, 'dummy') == 1



"""
Testing the DSRBF.
"""

from smote_variants import DSRBF

from smote_variants.datasets import (load_all_min_noise)

def test_specific():
    """
    Oversampler specific testing
    """

    obj = DSRBF(hidden_range=(30, 40))

    dataset = load_all_min_noise()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0



"""
This module implements the A_SUWO method.
"""

import numpy as np

from sklearn.cluster import AgglomerativeClustering
from sklearn.model_selection import KFold
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

from ..base import (NearestNeighborsWithMetricTensor,
                    pairwise_distances_mahalanobis, coalesce)
from ..base import OverSamplingSimplex
from ..base import coalesce_dict

from .._logger import logger
_logger= logger

__all__= ['A_SUWO']

class A_SUWO(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{a_suwo,
                        title = "Adaptive semi-unsupervised weighted
                                    oversampling (A-SUWO) for imbalanced
                                    datasets",
                        journal = "Expert Systems with Applications",
                        volume = "46",
                        pages = "405 - 416",
                        year = "2016",
                        issn = "0957-4174",
                        doi = "https://doi.org/10.1016/j.eswa.2015.10.031",
                        author = "Iman Nekooeimehr and Susana K. Lai-Yuen",
                        keywords = "Imbalanced dataset, Classification,
                                        Clustering, Oversampling"
                        }

    Notes:
        * Equation (7) misses a division by R_j.
        * It is not specified how to sample from clusters with 1 instances.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_density_based,
                  OverSamplingSimplex.cat_noise_removal,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_clus_maj=7,
                 c_thres=0.8,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling params
            n_clus_maj (int): number of majority clusters
            c_thres (float): threshold on distances
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                                'within_simplex_sampling': 'random',
                                'gaussian_component': None}

        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params,
                            random_state=random_state,
                            checks=None)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(n_clus_maj, "n_clus_maj", 1)
        self.check_greater_or_equal(c_thres, "c_thres", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_clus_maj = n_clus_maj
        self.c_thres = c_thres
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'n_clus_maj': [5, 7, 9],
                                  'c_thres': [0.5, 0.8]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def remove_noise(self, X, y, nn_params):
        """
        Noise removal

        Args:
            X (np.array): input vectors
            y (np.array): input labels
            nn_params (dict): nearest neighbors parameters

        Returns:
            np.array, np.array: the cleaned X and y
        """
        # fitting nearest neighbors to find neighbors of all samples
        n_neighbors = min([len(X), self.n_neighbors + 1])
        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **nn_params)
        nearestn.fit(X)
        ind = nearestn.kneighbors(X, return_distance=False)

        # identifying as noise those samples which do not have neighbors of
        # the same label
        def noise_func(idx):
            return np.sum(y[ind[idx][1:]] == y[idx]) == 0
        noise = np.where(np.array([noise_func(idx) for idx in range(len(X))]))[0]

        # removing noise
        X = np.delete(X, noise, axis=0)
        y = np.delete(y, noise, axis=0)

        return X, y

    def initialize_clusters(self, X_maj, X_min):
        """
        Initialize the clusters.

        Args:
            X_maj: majority samples
            X_min: minority samples

        Returns:
            list, list: the majority and minority clusters
        """
        n_clus_maj = min([len(X_maj), self.n_clus_maj])

        # clustering majority samples
        aggclus = AgglomerativeClustering(n_clusters=n_clus_maj)
        aggclus.fit(X_maj)
        maj_clusters = [np.where(aggclus.labels_ == i)[0]
                        for i in range(n_clus_maj)]

        # initialize minority clusters
        min_clusters = [np.array([i]) for i in range(len(X_min))]

        return maj_clusters, min_clusters

    def initialize_distances(self,
                                X_maj,
                                X_min,
                                nn_params,
                                *,
                                maj_clusters,
                                min_clusters):
        """
        Initialize the distances

        Args:
            X_maj (np.array): majority samples
            X_min (np.array): minority samples
            nn_params (dict): the nearest neighbors parameters
            maj_clusters (list): the majority clusters
            min_clusters (list): the minority clusters

        Returns:
            np.arary, np.array: the majority and minority distance matrices
        """
        # compute minority distance matrix of cluster
        dm_min = pairwise_distances_mahalanobis(X_min,
                                    tensor=nn_params['metric_tensor'])
        for idx in range(len(dm_min)):
            dm_min[idx, idx] = np.inf

        # compute distance matrix of minority and majority clusters
        dm_maj = np.zeros(shape=(len(X_min), len(maj_clusters)))
        for idx, _ in enumerate(X_min):
            for jdx, maj_cluster in enumerate(maj_clusters):
                pairwd = pairwise_distances_mahalanobis(X_maj[maj_cluster],
                                                        Y=X_min[min_clusters[idx]],
                                                        tensor=nn_params['metric_tensor'])
                dm_maj[idx, jdx] = np.min(pairwd)

        return dm_maj, dm_min

    def determine_threshold(self, X_min, nn_params):
        """
        Determines the threshold.

        Args:
            X_min (np.array): the minority samples
            nn_params (dict): the nearest neighbors parameters

        Returns:
            float: the threshold
        """
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=len(X_min),
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nearestn.fit(X_min)
        dist, _ = nearestn.kneighbors(X_min)
        d_med = np.median(dist, axis=1)
        return np.mean(d_med)*self.c_thres

    def fix_distribution(self, dist):
        """
        Fixing a distribution.

        Args:
            dist (iterable): the distribution

        Returns:
            np.array: the fixed distribution
        """
        dist = np.nan_to_num(dist, nan=0.0)
        if np.sum(dist) == 0:
            return np.repeat(1.0, len(dist))
        return dist

    def adaptive_sub_cluster_sizing(self, min_clusters, X_min, X_maj):
        """
        Adaptive sub cluster sizing

        Args:
            min_clusters (list): minority clusters
            X_min (np.array): the minority samples
            X_maj (np.array): the majority samples

        Returns:
            np.array: the minority cluster distribution
        """
        eps = []
        # going through all minority clusters
        for clus in min_clusters:
            # checking if cluster size is higher than 1
            if len(clus) > 1:
                kfold = KFold(min([len(clus), 5]),
                                random_state=self.random_state,
                                shuffle=True)
                preds = []
                # executing k-fold cross validation with linear discriminant
                # analysis
                X_c = X_min[clus]
                for train, test in kfold.split(X_c):
                    X_train = np.vstack([X_maj, X_c[train]])
                    y_train = np.hstack([np.repeat(self.maj_label, len(X_maj)),
                                    np.repeat(self.min_label, len(X_c[train]))])
                    lda = LinearDiscriminantAnalysis()
                    lda.fit(X_train, y_train)
                    preds.append(lda.predict(X_c[test]))
                preds = np.hstack(preds)
                # extracting error rate
                eps.append(np.sum(preds == self.maj_label)/len(preds))
            else:
                eps.append(1.0)

        # sampling distribution over clusters
        eps = self.fix_distribution(eps)

        return eps/np.sum(eps)

    def within_cluster(self, X_maj, X_min, nn_params, min_clusters):
        """
        Determine within cluster distributions

        Args:
            X_maj (np.array): the majority samples
            X_min (np.array): the minority samples
            nn_params (dict): the nearest neighbors parameters
            min_clusters (list): the minority clusters

        Returns:
            list, list: the within cluster distribution and neighbors
        """
        # synthetic instance generation - determining within cluster
        # distribution finding majority neighbor distances of minority
        # samples
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=1,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nearestn.fit(X_maj)
        dist, _ = nearestn.kneighbors(X_min)
        if np.all(dist == 0.0):
            dist[:, :] = 1.0
        else:
            dist[dist == 0.0] = np.min(dist[dist != 0.0])

        dist = dist / X_maj.shape[1]
        dist = 1.0 / dist

        # computing the THs
        THs = [] # pylint: disable=invalid-name
        for clus in min_clusters:
            THs.append(np.mean(dist[clus, 0]))

        # determining within cluster distributions
        within_cluster_dist = []
        for idx, clus in enumerate(min_clusters):
            Gamma = dist[clus, 0] # pylint: disable=invalid-name
            Gamma[Gamma > THs[idx]] = THs[idx]
            within_cluster_dist.append(Gamma / np.sum(Gamma))

        # extracting within cluster neighbors
        within_cluster_neighbors = []
        for clus in min_clusters:
            n_neighbors = min([len(clus), self.n_neighbors])
            nearestn = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
            nearestn.fit(X_min[clus])
            within_cluster_neighbors.append(nearestn.kneighbors(X_min[clus])[1])

        return within_cluster_dist, within_cluster_neighbors

    def generate_samples(self,
                        n_to_sample,
                        min_clusters,
                        min_cluster_dist,
                        *,
                        within_dist,
                        within_neigh,
                        X_min):
        """
        Generate samples

        Args:
            n_to_sample (int): the number of samples to generate
            min_clusters (list): minority clusters
            min_cluster_dist (list): minority cluster distribution
            withins (list, list): within cluster distribution and neighbors
            within_cluster_neighbors (list): the within cluster neighbors
            X_min (np.array): the minority samples

        Returns:
            list: the generated samples
        """
        # fixing within class distribution
        within_dist = [self.fix_distribution(w_dist) for w_dist in within_dist]
        within_dist = [w_dist/np.sum(w_dist) for w_dist in within_dist]

        # generate random cluster indices
        cluster_indices = self.random_state.choice(len(min_clusters),
                                                    n_to_sample,
                                                    p=min_cluster_dist)

        cluster_unique, cluster_count = np.unique(cluster_indices,
                                                    return_counts=True)

        samples = []

        for idx, cluster in enumerate(cluster_unique):
            #if len(min_clusters[cluster]) > self.n_dim - 1:
            samples.append(self.sample_simplex(X=X_min[min_clusters[cluster]],
                                                indices=within_neigh[cluster],
                                                n_to_sample=cluster_count[idx],
                                                base_weights=within_dist[cluster]))
            #else:
            #    samp = self.random_state.choice(np.arange(len(within_dist[cluster])),
            #                                    cluster_count[idx],
            #                                    p=within_dist[cluster])
            #    samples.append(X_min[min_clusters[cluster]][samp])

        return np.vstack(samples)

    def cluster_minority_samples(self,
                                min_clusters,
                                dm_maj,
                                dm_min,
                                T # pylint: disable=invalid-name
                                ):
        """
        Clustering of minority samples

        Args:
            min_clusters (list): the minority clusters
            dm_maj (np.array): the majority distance matrix
            dm_min (np.array): the minority distance matrix
            T (float): the threshold

        Returns:
            list: the minority clusters
        """
        # do the clustering of minority samples
        while True:
            # finding minimum distance between minority clusters
            pi = np.min(dm_min) # pylint: disable=invalid-name

            # if the minimum distance is higher than the threshold, stop
            if pi > T:
                break

            # find cluster pair of minimum distance
            min_dist_pair = np.where(dm_min == pi)
            min_i = min_dist_pair[0][0]
            min_j = min_dist_pair[1][0]

            # Step 3 - find majority clusters closer than pi
            A = np.where(np.logical_and(dm_maj[min_i] < pi, # pylint: disable=invalid-name
                                        dm_maj[min_j] < pi))[0]

            # Step 4 - checking if there is a majority cluster between the
            # minority ones
            if len(A) > 0:
                dm_min[min_i, min_j] = np.inf
                dm_min[min_j, min_i] = np.inf
            else:
                # Step 5
                # unifying minority clusters
                min_clusters[min_i] = np.hstack([min_clusters[min_i],
                                                 min_clusters[min_j]])
                # removing one of them
                #min_clusters = np.delete(min_clusters, [min_j], axis=0)
                del min_clusters[min_j]

                # updating the minority distance matrix
                dm_min[min_i] = np.min(np.vstack([dm_min[min_i],
                                                  dm_min[min_j]]), axis=0)
                dm_min[:, min_i] = dm_min[min_i]
                # removing jth row and column (merged in i)
                dm_min = np.delete(dm_min, min_j, axis=0)
                dm_min = np.delete(dm_min, min_j, axis=1)

                # fixing the diagonal elements
                for idx in range(len(dm_min)):
                    dm_min[idx, idx] = np.inf

                # updating the minority-majority distance matrix
                dm_maj[min_i] = np.min(np.vstack([dm_maj[min_i],
                                                  dm_maj[min_j]]), axis=0)
                dm_maj = np.delete(dm_maj, min_j, axis=0)

        return min_clusters

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "no need for sampling")

        X_orig, y_orig = X, y

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        X, y = self.remove_noise(X, y, nn_params)

        # extracting modified minority and majority datasets
        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        if len(X_min) == 0:
            _logger.info("%s: All minority samples removed as noise",
                            self.__class__.__name__)
            return X_orig.copy(), y_orig.copy()

        maj_clusters, min_clusters = self.initialize_clusters(X_maj, X_min)

        # I think this can never happen
        # if len(maj_clusters) == 0:
        #    _logger.info("%s: Number of majority clusters is null",
        #                self.__class__.__name__)
        #    return X_orig.copy(), y_orig.copy()

        dms = self.initialize_distances(X_maj,
                                        X_min,
                                        nn_params,
                                        maj_clusters=maj_clusters,
                                        min_clusters=min_clusters)

        min_clusters = self.cluster_minority_samples(min_clusters,
                                                *dms,
                                                self.determine_threshold(X_min,
                                                                    nn_params))

        # adaptive sub-cluster sizing
        min_cluster_dist = self.adaptive_sub_cluster_sizing(min_clusters,
                                                            X_min,
                                                            X_maj)

        within_dist, within_neigh = self.within_cluster(X_maj,
                                                        X_min,
                                                        nn_params,
                                                        min_clusters)

        # do the sampling
        return (np.vstack([X, np.vstack(self.generate_samples(n_to_sample,
                                                    min_clusters,
                                                    min_cluster_dist,
                                                    within_dist=within_dist,
                                                    within_neigh=within_neigh,
                                                    X_min=X_min))]),
                np.hstack([y, np.repeat(self.min_label, n_to_sample)]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_clus_maj': self.n_clus_maj,
                'c_thres': self.c_thres,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module sets the package level logger.
"""

import logging

__all__= ['logger']

# setting the _logger format
logger = logging.getLogger('smote_variants')
logger.setLevel(logging.DEBUG)
logger_ch = logging.StreamHandler()
logger_ch.setFormatter(logging.Formatter(
    "%(asctime)s:%(levelname)s:%(message)s"))
logger.addHandler(logger_ch)



"""
This module implements the SDSMOTE method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict, fix_density
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['SDSMOTE']

class SDSMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{sdsmote,
                            author={Li, K. and Zhang, W. and Lu, Q. and
                                        Fang, X.},
                            booktitle={2014 International Conference on
                                        Identification, Information and
                                        Knowledge in the Internet of
                                        Things},
                            title={An Improved SMOTE Imbalanced Data
                                    Classification Method Based on Support
                                    Degree},
                            year={2014},
                            volume={},
                            number={},
                            pages={34-38},
                            keywords={data mining;pattern classification;
                                        sampling methods;improved SMOTE
                                        imbalanced data classification
                                        method;support degree;data mining;
                                        class distribution;imbalanced
                                        data-set classification;over sampling
                                        method;minority class sample
                                        generation;minority class sample
                                        selection;minority class boundary
                                        sample identification;Classification
                                        algorithms;Training;Bagging;Computers;
                                        Testing;Algorithm design and analysis;
                                        Data mining;Imbalanced data-sets;
                                        Classification;Boundary sample;Support
                                        degree;SMOTE},
                            doi={10.1109/IIKI.2014.14},
                            ISSN={},
                            month={Oct}}
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling params
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def calculate_density(self, X_min, X_maj, nn_params):
        """
        Calculate the density

        Args:
            X_min (np.array): minority samples
            X_maj (np.array): majority samples
            nn_params (dict): nearest neighbor parameters

        Returns:
            np.array: the density
        """
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=len(X_maj),
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X_maj)
        dist, _ = nnmt.kneighbors(X_min)

        # calculating the sum according to S3 in the paper
        S_i = np.sum(dist, axis=1) # pylint: disable=invalid-name

        # calculating average distance according to S5
        S = np.sum(S_i) # pylint: disable=invalid-name
        S_ave = S/(len(X_min)*len(X_maj)) # pylint: disable=invalid-name

        k = np.array([len(nnmt.radius_neighbors(X_min[i].reshape(1, -1), # pylint: disable=invalid-name
                                                S_ave,
                                                return_distance=False))
                                        for i in range(len(X_min))])

        density = k/np.sum(k)

        return fix_density(density)


    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        # fitting nearest neighbors model to find closest majority points to
        # minority samples
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X, y)

        density = self.calculate_density(X_min, X_maj, nn_params)

        # fitting nearest neighbors model to minority samples to run
        # SMOTE-like sampling
        n_neighbors = min([len(X_min), self.n_neighbors+1])
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        samples = self.sample_simplex(X=X_min,
                                        indices=ind,
                                        n_to_sample=n_to_sample,
                                        base_weights=density)

        # do the sampling
        #samples = []
        #while len(samples) < n_to_sample:
        #    idx = self.random_state.choice(np.arange(len(density)), p=density)
        #    random_neighbor_idx = self.random_state.choice(ind[idx][1:])
        #    X_a = X_min[idx]
        #    X_b = X_min[random_neighbor_idx]
        #    samples.append(self.sample_between_points(X_a, X_b))

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the MWMOTE method.
"""
import warnings

import numpy as np

from sklearn.cluster import KMeans
from scipy.linalg import circulant

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['MWMOTE']

class MWMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @ARTICLE{mwmote,
                        author={Barua, S. and Islam, M. M. and Yao, X. and
                                Murase, K.},
                        journal={IEEE Transactions on Knowledge and Data
                                Engineering},
                        title={MWMOTE--Majority Weighted Minority Oversampling
                                Technique for Imbalanced Data Set Learning},
                        year={2014},
                        volume={26},
                        number={2},
                        pages={405-425},
                        keywords={learning (artificial intelligence);pattern
                                    clustering;sampling methods;AUC;area under
                                    curve;ROC;receiver operating curve;G-mean;
                                    geometric mean;minority class cluster;
                                    clustering approach;weighted informative
                                    minority class samples;Euclidean distance;
                                    hard-to-learn informative minority class
                                    samples;majority class;synthetic minority
                                    class samples;synthetic oversampling
                                    methods;imbalanced learning problems;
                                    imbalanced data set learning;
                                    MWMOTE-majority weighted minority
                                    oversampling technique;Sampling methods;
                                    Noise measurement;Boosting;Simulation;
                                    Complexity theory;Interpolation;Abstracts;
                                    Imbalanced learning;undersampling;
                                    oversampling;synthetic sample generation;
                                    clustering},
                        doi={10.1109/TKDE.2012.232},
                        ISSN={1041-4347},
                        month={Feb}}

    Notes:
        * The original method was not prepared for the case of having clusters
            of 1 elements.
        * It is not clear if within the cluster of the informative minority item
            the informative minority item should be choosen again
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 k1=5,
                 k2=5,
                 k3=5,
                 M=10,
                 cf_th=5.0,
                 cmax=10.0,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            k1 (int): parameter of the NearestNeighbors component
            k2 (int): parameter of the NearestNeighbors component
            k3 (int): parameter of the NearestNeighbors component
            M (int): number of clusters
            cf_th (float): cutoff threshold
            cmax (float): maximum closeness value
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(k1, 'k1', 1)
        self.check_greater_or_equal(k2, 'k2', 1)
        self.check_greater_or_equal(k3, 'k3', 1)
        self.check_greater_or_equal(M, 'M', 1)
        self.check_greater_or_equal(cf_th, 'cf_th', 0)
        self.check_greater_or_equal(cmax, 'cmax', 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.params = {'k1': k1,
                        'k2': k2,
                        'k3': k3,
                        'M': M,
                        'cf_th': cf_th,
                        'cmax': cmax}
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'k1': [5, 9],
                                  'k2': [5, 9],
                                  'k3': [5, 9],
                                  'M': [4, 10],
                                  'cf_th': [5.0],
                                  'cmax': [10.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def filter_minorities(self, X, y, nn_params, minority):
        """
        Filter minority samples.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            nn_params (dict): the nearest neighbors parameters
            minority (np.array): the minority indices

        Returns:
            np.array: the filtered minority indices
        """
        # Step 1
        n_neighbors = min([len(X), self.params['k1'] + 1])
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        ind1 = nnmt.kneighbors(X, return_distance=False)

        # Step 2
        arr = [i for i in minority if np.sum(y[ind1[i][1:]] == self.min_label)]
        filtered_minority = np.array(arr)

        return filtered_minority

    def determine_border_majority(self, X, X_maj,
                                    filtered_minority,
                                    nn_params):
        """
        Determine border majority samples.

        Args:
            X (np.array): all training samples
            X_maj (np.array): all majority samples
            filtered_minority (np.array): filtered minority samples
            nn_params (dict): nearest neighbors parameters

        Returns:
            np.array: the bordering majority samples
        """
        # Step 3 - ind2 needs to be indexed by indices of the lengh of X_maj
        n_neighbors = min([len(X_maj), self.params['k2'] + 1])
        nn_maj= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nn_maj.fit(X_maj)
        ind2 = nn_maj.kneighbors(X[filtered_minority], return_distance=False)

        # Step 4
        border_majority = np.unique(ind2.flatten())

        return border_majority

    def determine_informative_minority(self, X_min, X_maj,
                                        border_majority, nn_params):
        """
        Determine the informative minority samples.

        Args:
            X_min (np.array): minority samples
            X_maj (np.array): majority samples
            border_majority (np.array): border majority samples
            nn_params (dict): the nearest neighbors parameters

        Returns:
            np.array: indices of informative minority samples
        """

        # Step 5 - ind3 needs to be indexed by indices of the length of X_min
        n_neighbors = min([self.params['k3'], len(X_min)])
        nn_min = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nn_min.fit(X_min)
        ind3 = nn_min.kneighbors(X_maj[border_majority], return_distance=False)

        # Step 6 - informative minority indexes X_min
        informative_minority = np.unique(ind3.flatten())

        return informative_minority

    def closeness_factor(self, y_vec, x_vec):
        """
        Closeness factor according to the Eq (6)

        Args:
            y_vec (np.array): training instance (border_majority)
            x_vec (np.array): training instance (informative_minority)
            cf_th (float): cutoff threshold
            cmax (float): maximum values

        Returns:
            float: closeness factor
        """
        dist = np.linalg.norm(y_vec - x_vec) / y_vec.shape[0]

        #if dist == 0:
        #    dist = 0.1

        dist = np.max([0.1, dist])

        f_val = np.min([1.0 / dist, self.params['cf_th']])

        return f_val / self.params['cf_th'] * self.params['cmax']

    def determine_information_weights(self,
                                    X_maj,
                                    X_min,
                                    border_majority,
                                    informative_minority):
        """
        Determine information weights.

        Args:
            X_maj (np.array): majority samples
            X_min (np.array): minority samples
            border_majority (np.array): border majority indices
            informative minority (np.array): informative minority indices

        Returns:
            np.array: information weights
        """
        closeness_factors = np.zeros(shape=(border_majority.shape[0],
                                            informative_minority.shape[0]))

        for idx, bm_i in enumerate(border_majority):
            for jdx, im_j in enumerate(informative_minority):
                closeness_factors[idx, jdx] = self.closeness_factor(X_maj[bm_i],
                                                                    X_min[im_j])

        _logger.info("%s: computing information weights", self.__class__.__name__)
        information_weights = np.zeros(shape=(border_majority.shape[0],
                                              informative_minority.shape[0]))
        for idx in range(border_majority.shape[0]):
            norm_factor = np.sum(closeness_factors[idx, :])
            for jdx in range(informative_minority.shape[0]):
                cf_ij = closeness_factors[idx, jdx]
                information_weights[idx, jdx] = cf_ij**2 / norm_factor

        return information_weights

    def determine_selection_probabilities(self, *, X, y,
                                            X_min,
                                            filtered_minority,
                                            nn_params):
        """
        Determine selection probabilities.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            X_min (np.array): minority samples
            filtered_minority (np.array): indices of the filtered
                                            minority samples
            minority (np.array): the minority indices
            nn_params (dict): the nearest neighbors parameters

        Returns:
            np.array, np.array: indices of informative minority and
                                selection probabilities
        """
        X_maj = X[y == self.maj_label]

        border_majority = self.determine_border_majority(X, X_maj,
                                                            filtered_minority,
                                                            nn_params)

        informative_minority = self.determine_informative_minority(X_min, X_maj,
                                                    border_majority, nn_params)

        # Steps 7 - 9
        _logger.info("%s: computing closeness factors", self.__class__.__name__)

        information_weights = self.determine_information_weights(X_maj,
                                                        X_min,
                                                        border_majority,
                                                        informative_minority)

        selection_weights = np.sum(information_weights, axis=0)
        selection_probabilities = selection_weights/np.sum(selection_weights)

        return informative_minority, selection_probabilities

    def determine_cluster_probs(self, informative_minority,
                                    selection_probabilities,
                                    cluster_labels):
        """
        Determine cluster probabilities.

        Args:
            informative_minority (np.array): indices of informative minority
                                                sample indices
            selection_probabilities (np.array): selection probabilities
            cluster_labels (np.array): cluster labels

        Returns:
            np.array: the cluster probabilities
        """
        cluster_weights = np.zeros(shape=(np.max(cluster_labels) + 1,))
        for idx, p_idx in enumerate(informative_minority):
            cluster_weights[cluster_labels[p_idx]] += \
                                    selection_probabilities[idx]

        return cluster_weights / np.sum(cluster_weights)

    def determine_within_prob(self,
                                cluster,
                                informative_minority,
                                selection_probabilities):
        """
        Determine the within cluster probabilities.

        Args:
            cluster (np.array): the cluster indices
            informative_minority (np.array): the informative minority
                                                indices
            selection_probabilities (np.array): the informative minority
                                                probabilities

        Returns:
            np.array: the within cluster probabilities
        """
        within_prob = np.zeros(shape=(len(cluster),))

        for idx, c_idx in enumerate(cluster):
            where = np.where(informative_minority == c_idx)[0]
            if len(where) > 0:
                within_prob[idx] += selection_probabilities[where[0]]

        within_prob = within_prob / np.sum(within_prob)

        return within_prob

    def sample_clusters(self,
                        informative_minority,
                        selection_probabilities,
                        cluster_labels,
                        n_to_sample):
        """
        Sample the clusters.

        Args:
            informative_minority (np.array): indices of the informative
                                                minority
            selection_probabilities (np.array): the selection probabilities
            cluster_labels (np.array): the cluster labels
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array, np.array: the selected cluster labels and counts
        """
        cluster_probs = self.determine_cluster_probs(informative_minority,
                                                    selection_probabilities,
                                                    cluster_labels)

        clusters_selected = self.random_state.choice(len(cluster_probs),
                                                        n_to_sample,
                                                        p=cluster_probs)

        cluster_unique, cluster_count = np.unique(clusters_selected,
                                                    return_counts=True)

        return cluster_unique, cluster_count

    def generate_samples_in_clusters(self, *, X_min, clusters,
                                    cluster_labels,
                                    informative_minority,
                                    selection_probabilities,
                                    n_to_sample):
        """
        Generate samples in clusters.

        Args:
            X_min (np.array): the minority vectors
            clusters (np.array): the clusters
            cluster_labels (np.array): the cluster labeling
            informative_minority (np.array): indices of the informative
                                                minority
            selection_probabilities (np.array): the selection probabilities
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        # Step 11-12
        cluster_unique, cluster_count = \
                self.sample_clusters(informative_minority,
                                    selection_probabilities,
                                    cluster_labels,
                                    n_to_sample)

        #n_dim_original = self.n_dim
        samples = []
        for idx, cluster in enumerate(cluster_unique):
            cluster_vectors = X_min[clusters[cluster]]
            within_prob = self.determine_within_prob(clusters[cluster],
                                                    informative_minority,
                                                    selection_probabilities)

            #self.n_dim = np.min([self.n_dim, cluster_vectors.shape[0]])
            samples.append(self.sample_simplex(X=cluster_vectors,
                                    indices=circulant(np.arange(cluster_vectors.shape[0])),
                                    n_to_sample=cluster_count[idx],
                                    base_weights=within_prob))
            #self.n_dim = n_dim_original
        #"""


        #for cluster in clusters:
        #    if len(cluster) == 1:
        #        print(X_min[cluster])
        #
        #samples=[]
        ## Step 12
        #
        #for _ in range(n_to_sample):
        #    random_index = self.random_state.choice(informative_minority,
        #                                            p=selection_probabilities)
        #    cluster_label = cluster_labels[random_index]
        #    cluster = clusters[cluster_label]
        #    random_index_in_cluster = self.random_state.choice(cluster)
        #    print(cluster_label, random_index, random_index_in_cluster)
        #    X_random = X_min[random_index]
        #    X_random_cluster = X_min[random_index_in_cluster]
        #    #samples.append(self.sample_between_points(X_random,
        #    #                                          X_random_cluster))
        #    samples.append(X_random + (X_random_cluster - X_random)\
        # *self.random_state.random_sample())

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)
        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X, y)

        filtered_minority = self.filter_minorities(X, y, nn_params,
                                            np.where(y == self.min_label)[0])

        if len(filtered_minority) == 0:
            return self.return_copies(X, y,
                        "No minority samples remaining after filtering")

        informative_minority, selection_probabilities = \
                self.determine_selection_probabilities(X=X, y=y,
                                                        X_min=X_min,
                                                        filtered_minority=filtered_minority,
                                                        nn_params=nn_params)

        # Step 10
        _logger.info("%s: do clustering", self.__class__.__name__)
        kmeans = KMeans(n_clusters=np.min([len(np.unique(X_min, axis=1)),
                                            self.params['M']]),
                        random_state=self._random_state_init)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            kmeans.fit(X_min)
        #imin_labels = kmeans.labels_[informative_minority]

        clusters = [np.where(kmeans.labels_ == i)[0]
                    for i in range(np.max(kmeans.labels_)+1)]

        samples = self.generate_samples_in_clusters(X_min=X_min,
                            clusters=clusters,
                            cluster_labels=kmeans.labels_,
                            informative_minority=informative_minority,
                            selection_probabilities=selection_probabilities,
                            n_to_sample=n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'k1': self.params['k1'],
                'k2': self.params['k2'],
                'k3': self.params['k3'],
                'M': self.params['M'],
                'cf_th': self.params['cf_th'],
                'cmax': self.params['cmax'],
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing oversampling with 1 minority sample only.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import load_1_min_some_maj
from .additional_objs import additional_objs

dataset = load_1_min_some_maj()

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_1_min_some_maj(smote_class):
    """
    Testing oversamplers with 1 minority sample.

    Args:
        smote_class (class): an oversampler class.
    """
    X, y = smote_class().sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", additional_objs)
def test_1_min_some_maj_additional(smote_obj):
    """
    Testing an oversampler with 1 minority sample.

    Args:
        smote_obj (obj): an oversampler obj
    """
    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
This module implements the Supervised_SMOTE method.
"""

import numpy as np

from scipy.linalg import circulant

from ..base import OverSamplingSimplex
from ..base import instantiate_obj, coalesce_dict

from .._logger import logger
_logger= logger

__all__= ['Supervised_SMOTE']

class Supervised_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{supervised_smote,
                        author = {Hu, Jun AND He, Xue AND Yu, Dong-Jun AND
                                    Yang, Xi-Bei AND Yang, Jing-Yu AND Shen,
                                    Hong-Bin},
                        journal = {PLOS ONE},
                        publisher = {Public Library of Science},
                        title = {A New Supervised Over-Sampling Algorithm
                                    with Application to Protein-Nucleotide
                                    Binding Residue Prediction},
                        year = {2014},
                        month = {09},
                        volume = {9},
                        url = {https://doi.org/10.1371/journal.pone.0107676},
                        pages = {1-10},
                        number = {9},
                        doi = {10.1371/journal.pone.0107676}
                    }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_uses_classifier,
                  OverSamplingSimplex.cat_application]

    def __init__(self,
                 proportion=1.0,
                 *,
                 th_lower=0.5,
                 th_upper=1.0,
                 classifier=('sklearn.ensemble',
                                'RandomForestClassifier',
                                {'n_estimators': 50,
                                'n_jobs': 1,
                                'random_state': 5}),
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            th_lower (float): lower bound of the confidence interval
            th_upper (float): upper bound of the confidence interval
            classifier (obj): classifier used to estimate class memberships
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                    'within_simplex_sampling': 'random',
                    'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_in_range(th_lower, "th_lower", [0, 1])
        self.check_in_range(th_upper, "th_upper", [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.th_lower = th_lower
        self.th_upper = th_upper
        self.classifier = classifier
        self.classifier_obj = instantiate_obj(classifier)
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        classifiers = [('sklearn.ensemble',
                        'RandomForestClassifier',
                        {'n_estimators': 50,
                        'n_jobs': 1,
                        'random_state': 5})]
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'th_lower': [0.3, 0.5, 0.8],
                                  'th_upper': [1.0],
                                  'classifier': classifiers}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # training the classifier
        self.classifier_obj.fit(X, y)

        X_min = X[y == self.min_label]

        th_lower = self.th_lower

        indices = circulant(np.arange(X_min.shape[0]))

        class_column = np.where(self.classifier_obj.classes_ == self.min_label)
        class_column = class_column[0][0]

        samples = np.zeros(shape=(0, X.shape[1]))
        n_trials = 1
        n_success = 1
        while len(samples) < n_to_sample:
            to_sample = np.max([n_to_sample - len(samples), 10])

            n_trials = n_trials + to_sample

            samples_tmp = self.sample_simplex(X=X_min,
                                          indices=indices,
                                          n_to_sample=to_sample)

            prob = self.classifier_obj.predict_proba(samples_tmp)
            prob = prob[:, class_column]

            samples_tmp = samples_tmp[(prob >= th_lower)
                                    & (prob <= self.th_upper)]

            samples = np.vstack([samples, samples_tmp])

            n_success = n_success + samples_tmp.shape[0]

            if n_success/n_trials < 0.02:
                th_lower = th_lower * 0.9
                n_success = 1
                n_trials = 1

        samples = samples[:np.min([samples.shape[0], n_to_sample])]

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'th_lower': self.th_lower,
                'th_upper': self.th_upper,
                'classifier': self.classifier,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing the ADG.
"""
import numpy as np

import pytest

from smote_variants import ADG

from smote_variants.datasets import load_normal

def test_specific():
    """
    Oversampler specific testing
    """
    with pytest.raises(Exception):
        ADG(kernel='dummy')

    with pytest.raises(Exception):
        ADG(kernel='rbf_-6')

    dataset = load_normal()

    X_samp, _ = ADG(kernel='rbf_1').sample(dataset['data'],
                                           dataset['target'])

    assert len(X_samp) > 0

    X_samp, _ = ADG().sample(dataset['data'],
                             dataset['target'])

    assert len(X_samp) > 0

    with pytest.raises(Exception):
        ADG().check_early_stopping(0.0)

    with pytest.raises(Exception):
        ADG().step_9_16(partial_results={'Z_hat': np.array([[]])},
                        kernel_function='inner',
                        y=None)



"""
Testing oversampling with linearly dependent data.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import load_linearly_dependent
from .additional_objs import additional_objs

dataset = load_linearly_dependent()

@pytest.mark.parametrize("smote_class", sv.get_all_oversamplers())
def test_linearly_dependent(smote_class):
    """
    Testing oversamplers with 2 minority samples.

    Args:
        smote_class (class): an oversampler class.
    """
    X, y = smote_class().sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0

@pytest.mark.parametrize("smote_obj", additional_objs)
def test_linearly_dependent_additional(smote_obj):
    """
    Testing oversamplers with 2 minority samples.

    Args:
        smote_obj (obj): an oversampler obj
    """
    X, y = smote_obj.sample(dataset['data'], dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
Testing the base functionalities.
"""

import os
import shutil
import json

import numpy as np
from sklearn.tree import DecisionTreeClassifier

import pytest

from smote_variants.base import (mode, StatisticsMixin, RandomStateMixin,
                            ParametersMixin, instantiate_obj, coalesce,
                            load_dict, dump_dict, check_if_damaged,
                            equal_dicts, coalesce_dict, safe_divide,
                            fix_density, cov)

def test_mode():
    """
    Testing the mode function.
    """
    assert mode([0, 1, 1, 2]) == 1

def test_cov():
    """
    Test the cov function.
    """
    assert len(cov(np.array([[1], [2], [3]]), rowvar=False).shape) == 2
    assert len(cov(np.array([[1, 2, 3]]), rowvar=True).shape) == 2

    assert cov(np.array([[1, 2], [2, 3]]), rowvar=False).shape[0] == 2
    assert cov(np.array([[1, 2], [2, 3]]), rowvar=True).shape[0] == 2

def test_fix_density():
    """
    Test the fix density function
    """
    assert fix_density(np.array([1.0, 1.0]))[0] == 0.5
    assert fix_density(np.array([1.0, np.nan]))[0] == 1.0
    assert fix_density(np.array([np.nan, np.inf]))[0] == 0.5

def test_safe_divide():
    """
    Test the safe_divide function
    """
    assert safe_divide(1, 2, 0.0) == 1.0/2.0
    assert safe_divide(1, 0, 5) == 5

def test_equal_dicts():
    """
    Testing the equality check of dictionaries.
    """
    assert equal_dicts({'a': 'a'}, {'a': 'a'})
    assert equal_dicts({'a': {'a': 'a'}}, {'a': {'a': 'a'}})
    assert not equal_dicts({'a': 'a'}, {'a': 'b'})
    assert not equal_dicts({'a': {'a': 'a'}}, {'a': {'a': 'b'}})
    assert not equal_dicts({'a': {'a': 'a'}}, {'a': {'a': 'b'}, 'b': 2})

def test_coalesce():
    """
    Testing the coalesce function
    """
    assert coalesce(None, 1) == 1
    assert coalesce(2, 1) == 2

def test_coalesce_dict():
    """
    Testing the coalesce dict functionality
    """
    dict_1 = {'a': 0, 'b': 1, 'd': 4}
    dict_0 = {'b': 2, 'c': 3}

    result = coalesce_dict(dict_0, dict_1)

    assert len(result) == 4
    assert result['b'] == 2

    result = coalesce_dict(None, dict_1)

    assert equal_dicts(dict_1, result)

def test_instantiate_obj():
    """
    Testing the object instantiation
    """
    obj = instantiate_obj(('sklearn.tree', 'DecisionTreeClassifier', {}))
    assert obj.__class__ == DecisionTreeClassifier

def test_load_dump_dict():
    """
    Testing the dictionary dumping
    """
    test_dir = 'test_base_dir'
    os.makedirs(test_dir, exist_ok=True)

    filename = os.path.join(test_dir, 'dump.pickle')

    obj = {'a': 2,
            'b': np.array([1, 2])}
    dump_dict(obj, filename, 'pickle', ['b'])
    obj_loaded = load_dict(filename, 'pickle', ['b'])

    np.testing.assert_array_equal(obj['b'], obj_loaded['b'])

    dump_dict(obj, filename, 'json', ['b'])
    obj_loaded = load_dict(filename, 'json', ['b'])

    np.testing.assert_array_equal(obj['b'], obj_loaded['b'])

    with pytest.raises(ValueError):
        dump_dict(obj, filename, 'nonsense', ['b'])

    with pytest.raises(ValueError):
        load_dict(filename, 'nonsense', ['b'])

    obj = {'a': 2}
    dump_dict(obj, filename, 'pickle')
    obj_loaded = load_dict(filename, 'pickle')

    assert obj['a'] == obj_loaded['a']

    shutil.rmtree(test_dir)

def test_check_if_damaged():
    """
    Testing the file availability functionality.
    """
    test_dir = 'test_base_dir'
    os.makedirs(test_dir, exist_ok=True)

    assert check_if_damaged(os.path.join(test_dir, 'nonsense.json'), 'json')

    with open(os.path.join(test_dir, 'a.json'), 'wt', encoding='utf-8') as file:
        file.write('nonsense')

    assert check_if_damaged(os.path.join(test_dir, 'a.json'), 'json')

    with open(os.path.join(test_dir, 'a.json'), 'wt', encoding='utf-8') as file:
        json.dump({'a': 2}, file)

    assert not check_if_damaged(os.path.join(test_dir, 'a.json'), 'json')

    shutil.rmtree(test_dir)


def test_statistics_mixin():
    """
    Testing the statistics mixin.
    """

    class Mock(StatisticsMixin):
        """
        Mock class
        """

    mock = Mock()

    mock.class_label_statistics([0, 0, 0, 1])

    assert mock.min_label == 1
    assert mock.maj_label == 0

    assert len(StatisticsMixin().get_params()) == 0

def test_random_state_mixin():
    """
    Testing the random state mixin.
    """
    obj = RandomStateMixin(random_state=None)
    assert obj.random_state == np.random

    obj = RandomStateMixin(random_state=5)
    assert isinstance(obj.random_state, np.random.RandomState)
    assert obj.get_params()['random_state'] == 5

    obj = RandomStateMixin(random_state=np.random.RandomState(5))
    assert isinstance(obj.random_state, np.random.RandomState)

    obj = RandomStateMixin(random_state=np.random)
    assert obj.random_state == np.random

    with pytest.raises(ValueError):
        obj = RandomStateMixin(random_state='apple')

    assert len(obj.get_params()) == 1

def test_parameters_mixin():
    """
    Testing the parameters mixin
    """
    pmixin = ParametersMixin()

    with pytest.raises(ValueError):
        pmixin.check_in_range(5, 'five', [0, 2])

    with pytest.raises(ValueError):
        pmixin.check_out_range(1, 'five', [0, 2])

    with pytest.raises(ValueError):
        pmixin.check_less_or_equal(5, 'five', 3)

    with pytest.raises(ValueError):
        pmixin.check_less_or_equal_par(5, 'five', 3, 'three')

    with pytest.raises(ValueError):
        pmixin.check_less(5, 'five', 3)

    with pytest.raises(ValueError):
        pmixin.check_less_par(5, 'five', 3, 'three')

    with pytest.raises(ValueError):
        pmixin.check_greater_or_equal(5, 'five', 8)

    with pytest.raises(ValueError):
        pmixin.check_greater_or_equal_par(5, 'five', 8, 'three')

    with pytest.raises(ValueError):
        pmixin.check_greater(5, 'five', 8)

    with pytest.raises(ValueError):
        pmixin.check_greater_par(5, 'five', 8, 'three')

    with pytest.raises(ValueError):
        pmixin.check_equal(5, 'five', 5)

    with pytest.raises(ValueError):
        pmixin.check_equal_par(5, 'five', 5, 'three')

    with pytest.raises(ValueError):
        pmixin.check_isin(5, 'five', [1, 2, 6])

    with pytest.raises(ValueError):
        pmixin.check_n_jobs(None)

def test_parameter_combinations():
    """
    Testing the parameter combinations functionality.
    """
    params_base = {'a': [1, 2],
                    'b': ['a', 'b', 'c']}
    params = ParametersMixin.generate_parameter_combinations(params_base,
                                                                raw=False)
    assert len(params) == 6

    assert len(ParametersMixin.generate_parameter_combinations(params_base,
                                                                raw=True)) == 2



"""
This module implements the NRSBoundary_SMOTE method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import (NearestNeighborsWithMetricTensor,
                                pairwise_distances_mahalanobis)
from ..base import OverSamplingSimplex

from .._logger import logger
_logger= logger

__all__= ['NRSBoundary_SMOTE']

class NRSBoundary_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @Article{nrsboundary_smote,
                    author= {Feng, Hu and Hang, Li},
                    title= {A Novel Boundary Oversampling Algorithm Based on
                            Neighborhood Rough Set Model: NRSBoundary-SMOTE},
                    journal= {Mathematical Problems in Engineering},
                    year= {2013},
                    pages= {10},
                    doi= {10.1155/2013/694809},
                    url= {http://dx.doi.org/10.1155/694809}
                    }
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 w=0.005, # pylint: disable=invalid-name
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors in nearest neighbors
                                component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            w (float): used to set neighborhood radius
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(w, "w", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.w = w # pylint: disable=invalid-name
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'w': [0.005, 0.01, 0.05]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_delta(self, X, nn_params):
        """
        Determine the delta values

        Args:
            X (np.array): all training vectors
            nn_params (dict): nearest neighbor parameters

        Returns:
            np.array: the delta values
        """
        distm = pairwise_distances_mahalanobis(X,
                                               tensor=nn_params['metric_tensor'])
        d_max = np.max(distm, axis=1)
        np.fill_diagonal(distm, np.max(distm))
        d_min = np.min(distm, axis=1)

        delta = d_min + self.w*(d_max - d_min)

        return delta

    def determine_bound_pos_set_delta(self, X, y, nn_params):
        """
        Determine bound_set and pos_set and the delta vector.

        Args:
            X (np.array): training vectors
            y (np.array): target labels
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array, np.array: the bound_set, pos_set and delta
        """
        delta = self.determine_delta(X, nn_params)

        # step 1
        bound_set = []
        pos_set = []

        # number of neighbors is not interesting here, as we use the
        # radius_neighbors function to extract the neighbors in a given radius
        n_neighbors = np.min([self.n_neighbors + 1, len(X)])

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        for idx, x_vec in enumerate(X):
            indices = nnmt.radius_neighbors(x_vec.reshape(1, -1),
                                          delta[idx],
                                          return_distance=False)

            n_minority = np.sum(y[indices[0].astype(int)] == self.min_label)
            n_majority = np.sum(y[indices[0].astype(int)] == self.maj_label)
            if y[idx] == self.min_label and not n_minority == len(indices[0]):
                bound_set.append(idx)
            elif y[idx] == self.maj_label and n_majority == len(indices[0]):
                pos_set.append(idx)

        bound_set = np.array(bound_set)
        pos_set = np.array(pos_set)

        return bound_set, pos_set, delta

    def generate_samples(self, *, X, X_min,
                    indices, bound_set, pos_set,
                    delta, n_to_sample):
        """
        Generate samples.

        Args:
            X (np.array): all training vectors
            X_min (np.array): minority vectors
            indices (np.array): the neighborhood structure
            bound_set (np.array): indices of the boundary set
            pos_set (np.array): indices of the positive set
            delta (np.array): delta distances
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the new samples
        """
        trials = 0
        samples = np.zeros(shape=(0, X.shape[1]))

        while len(samples) < n_to_sample and trials < n_to_sample:
            subsample = self.sample_simplex(X=X[bound_set],
                                            indices=indices,
                                            n_to_sample=(n_to_sample - len(samples))*4,
                                            X_vertices=X_min)
            # checking the conflict
            dist_from_pos_set = \
                    np.linalg.norm(subsample - X[pos_set][:, None], axis=2)
            no_conflict = np.all(dist_from_pos_set.T > delta[pos_set], axis=1)
            samples = np.vstack([samples, subsample[no_conflict]])

            trials = trials + 1

        return np.vstack(samples)[:n_to_sample]

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # determining the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        # step 2
        X_min_indices = np.where(y == self.min_label)[0] # pylint: disable=invalid-name
        X_min = X[X_min_indices]

        # step 3
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                    self.metric_tensor_from_nn_params(nn_params, X, y)

        bound_set, pos_set, delta = \
                    self.determine_bound_pos_set_delta(X, y, nn_params)

        if len(pos_set) == 0 or len(bound_set) == 0:
            return self.return_copies(X, y, "bound set or pos set empty")

        # step 4 and 5
        # computing the nearest neighbors of the bound set from the
        # minority set
        n_neighbors = min([len(X_min), self.n_neighbors + 1])
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        indices = nnmt.kneighbors(X[bound_set], return_distance=False)

        samples = self.generate_samples(X=X, X_min=X_min,
                        indices=indices, bound_set=bound_set, pos_set=pos_set,
                        delta=delta, n_to_sample=n_to_sample)


        # do the sampling
        #samples = []
        #trials = 0
        #w = self.w
        #while len(samples) < n_to_sample:
        #    idx = self.random_state.choice(len(bound_set))
        #    random_neighbor_idx = self.random_state.choice(indices[idx][1:])
        #    x_new = self.sample_between_points(
        #        X[bound_set[idx]], X_min[random_neighbor_idx])
        #
        #    # checking the conflict
        #    dist_from_pos_set = np.linalg.norm(X[pos_set] - x_new, axis=1)
        #    if np.all(dist_from_pos_set > delta[pos_set]):
        #        # no conflict
        #        samples.append(x_new)
        #    trials = trials + 1
        #    if trials > 1000 and len(samples) == 0:
        #        trials = 0
        #        w = w*0.9

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'w': self.w,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the LLE_SMOTE method.
"""

import numpy as np
from numpy.linalg import LinAlgError

from sklearn.manifold import LocallyLinearEmbedding

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex

from .._logger import logger
_logger= logger

__all__= ['LLE_SMOTE']

class LLE_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{lle_smote,
                            author={Wang, J. and Xu, M. and Wang,
                                    H. and Zhang, J.},
                            booktitle={2006 8th international Conference
                                    on Signal Processing},
                            title={Classification of Imbalanced Data by Using
                                    the SMOTE Algorithm and Locally Linear
                                    Embedding},
                            year={2006},
                            volume={3},
                            number={},
                            pages={},
                            keywords={artificial intelligence;
                                        biomedical imaging;medical computing;
                                        imbalanced data classification;
                                        SMOTE algorithm;
                                        locally linear embedding;
                                        medical imaging intelligence;
                                        synthetic minority oversampling
                                        technique;
                                        high-dimensional data;
                                        low-dimensional space;
                                        Biomedical imaging;
                                        Back;Training data;
                                        Data mining;Biomedical engineering;
                                        Research and development;
                                        Electronic mail;Pattern recognition;
                                        Performance analysis;
                                        Classification algorithms},
                            doi={10.1109/ICOSP.2006.345752},
                            ISSN={2164-5221},
                            month={Nov}}

    Notes:
        * There might be numerical issues if the nearest neighbors contain
            some element multiple times.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_dim_reduction,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 n_components=2,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj
                                and n_min to sample e.g. 1.0 means that after
                                sampling the number of minority samples will
                                be equal to the number of majority samples
            n_neighbors (int): control parameter of the nearest neighbor
                                component
            n_components (int): dimensionality of the embedding space
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 2)
        self.check_greater_or_equal(n_components, 'n_components', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.n_components = n_components
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'n_components': [2, 3, 5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def solve_for_weights(self,
                            xi, # pylint: disable=invalid-name
                            Z): # pylint: disable=invalid-name
        """
        Solve for locally linear embedding weights

        Args:
            xi (np.array): vector
            Z (np.array): matrix of neighbors in rows

        Returns:
            np.array: reconstruction weights

        Following https://cs.nyu.edu/~roweis/lle/algorithm.html
        """
        Z = Z - xi # pylint: disable=invalid-name
        Z = Z.T # pylint: disable=invalid-name
        C = np.dot(Z.T, Z) # pylint: disable=invalid-name
        try:
            weights = np.linalg.solve(C, np.repeat(1.0, C.shape[0]))
            if np.linalg.norm(weights) > 1e8:
                weights = np.repeat(1.0, C.shape[0])
        except LinAlgError:
            weights = np.repeat(1.0, C.shape[0])

        return weights/np.sum(weights)

    def embed_in_original_space(self,
                                X_min_transformed, # pylint: disable=invalid-name
                                X_min,
                                nnmt,
                                samples_raw):
        """
        Carry out the embedding in the original feature space

        Args:
            X_min_transformed (np.array): the minority samples after LLE
            X_min (np.array): the original minority samples
            nnmt (obj): fitted nearest neighbors object
            samples_raw (np.array): the raw samples

        Returns:
            np.array: the final samples
        """
        ind = nnmt.kneighbors(samples_raw, return_distance=False)

        samples = []
        for idx, sample in enumerate(samples_raw):
            weights = self.solve_for_weights(sample, X_min_transformed[ind[idx]])
            samples.append(np.dot(weights, X_min[ind[idx]]))

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        # determine the number of samples to generate
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # extracting minority samples
        X_min = X[y == self.min_label]

        n_components = np.min([self.n_components, X.shape[1]])
        n_neighbors = np.min([self.n_neighbors, X_min.shape[0] - 1])

        # do the locally linear embedding
        lle = LocallyLinearEmbedding(n_neighbors=n_neighbors,
                                        n_components=n_components,
                                        n_jobs=self.n_jobs)
        lle.fit(X_min)
        X_min_transformed = lle.transform(X_min) # pylint: disable=invalid-name

        # fitting the nearest neighbors model for sampling
        n_neighbors = min([self.n_neighbors + 1, len(X_min_transformed)])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params,
                                                                        lle.transform(X),
                                                                        y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min_transformed)
        ind = nnmt.kneighbors(X_min_transformed, return_distance=False)

        samples_raw = self.sample_simplex(X=X_min_transformed,
                                        indices=ind,
                                        n_to_sample=n_to_sample)

        samples = self.embed_in_original_space(X_min_transformed, # pylint: disable=invalid-name
                                                X_min,
                                                nnmt,
                                                samples_raw)

        # generating samples
        #samples = []
        #for _ in range(n_to_sample):
        #    idx = self.random_state.randint(len(X_min))
        #    random_coords = self.random_state.choice(ind[idx][1:])
        #    xi = self.sample_between_points(X_min_transformed[idx],
        #                                    X_min_transformed[random_coords])
        #    Z = X_min_transformed[ind[idx][1:]]
        #    w = solve_for_weights(xi, Z)
        #    samples.append(np.dot(w, X_min[ind[idx][1:]]))

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'n_components': self.n_components,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing the LN_SMOTE.
"""
import numpy as np

from smote_variants import LN_SMOTE

X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9]])
y = np.array([0, 1, 0, 1, 0, 1, 0])

def test_specific():
    """
    Oversampler specific testing
    """

    obj = LN_SMOTE()

    assert obj.random_gap(slp=1,
                            sln=0,
                            n_label=None,
                            n_neighbors=None,
                            n_dim=2).shape[0] == 2

    obj = LN_SMOTE(n_neighbors=2)

    X_samp, _ = obj.sample(X, y)

    assert len(X_samp) > 0

    obj = LN_SMOTE(n_neighbors=1)

    X_samp, _ = obj.sample(X, y)

    assert len(X_samp) > 0

    samples = obj.generate_samples(slpsln={},
                                    n_to_sample=5,
                                    X=np.array([[1, 2]]),
                                    y=None,
                                    n_neighbors=5)

    assert samples.shape[0] == 0



"""
Test the datasets module.
"""

from smote_variants.datasets import (load_1_dim,
                                     load_illustration_2_class,
                                     load_illustration_3_class,
                                     load_illustration_4_class,
                                     load_normal,
                                     load_same_num,
                                     load_some_min_some_maj,
                                     load_1_min_some_maj,
                                     load_2_min_some_maj,
                                     load_3_min_some_maj,
                                     load_4_min_some_maj,
                                     load_5_min_some_maj,
                                     load_1_min_1_maj,
                                     load_repeated,
                                     load_all_min_noise,
                                     load_separable,
                                     load_linearly_dependent,
                                     load_alternating,
                                     load_high_dim)

def test_dataset_loaders():
    """
    Testing dataset loaders
    """

    for loader in [load_1_dim,
                    load_illustration_2_class,
                    load_illustration_3_class,
                    load_illustration_4_class,
                    load_normal,
                    load_same_num,
                    load_some_min_some_maj,
                    load_1_min_some_maj,
                    load_2_min_some_maj,
                    load_3_min_some_maj,
                    load_4_min_some_maj,
                    load_5_min_some_maj,
                    load_1_min_1_maj,
                    load_repeated,
                    load_all_min_noise,
                    load_separable,
                    load_linearly_dependent,
                    load_alternating,
                    load_high_dim]:
        dataset = loader()
        assert len(dataset) == 3



"""
This module brings together all noise removal functionalities.
"""

from ._cnntomeklinks import *
from ._condensednn import *
from ._editednn import *
from ._neighborhoodcleaningrule import *
from ._noisefilter import *
from ._onesidedselection import *
from ._tomeklinkremoval import *



"""
This module implements the NEATER method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base._simplexsampling import array_array_index
from ..base import (NearestNeighborsWithMetricTensor,
                                pairwise_distances_mahalanobis)
from ..base import OverSampling
from ._smote import SMOTE
from ._adasyn import ADASYN

from .._logger import logger
_logger= logger

__all__= ['NEATER']

class NEATER(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{neater,
                            author={Almogahed, B. A. and Kakadiaris, I. A.},
                            booktitle={2014 22nd International Conference on
                                         Pattern Recognition},
                            title={NEATER: Filtering of Over-sampled Data
                                    Using Non-cooperative Game Theory},
                            year={2014},
                            volume={},
                            number={},
                            pages={1371-1376},
                            keywords={data handling;game theory;information
                                        filtering;NEATER;imbalanced data
                                        problem;synthetic data;filtering of
                                        over-sampled data using non-cooperative
                                        game theory;Games;Game theory;Vectors;
                                        Sociology;Statistics;Silicon;
                                        Mathematical model},
                            doi={10.1109/ICPR.2014.245},
                            ISSN={1051-4651},
                            month={Aug}}

    Notes:
        * Evolving both majority and minority probabilities as nothing ensures
            that the probabilities remain in the range [0,1], and they need to
            be normalized.
        * The inversely weighted function needs to be cut at some value (like
            the alpha level), otherwise it will overemphasize the utility of
            having differing neighbors next to each other.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_borderline,
                  OverSampling.cat_changes_majority,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 smote_n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 b=5,
                 alpha=0.1,
                 h=20,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            smote_n_neighbors (int): number of neighbors in SMOTE sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): the simplex sampling parameters
            b (int): number of neighbors
            alpha (float): smoothing term
            h (int): number of iterations in evolution
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(smote_n_neighbors, "smote_n_neighbors", 1)
        self.check_greater_or_equal(b, "b", 1)
        self.check_greater_or_equal(alpha, "alpha", 0)
        self.check_greater_or_equal(h, "h", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.smote_n_neighbors = smote_n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = ss_params
        self.params = {'b': b,
                        'alpha': alpha,
                        'h': h}
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'smote_n_neighbors': [3, 5, 7],
                                  'b': [3, 5, 7],
                                  'alpha': [0.1],
                                  'h': [20]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    #def wprob_mixed(self, prob, i, indices, distm, offset):
    #    ind = indices[i][1:]
    #    term_0 = 1*prob[offset + i][0]*prob[ind, 0]
    #    term_1 = distm[i, ind]*(prob[offset + i][1]*prob[ind, 0] +
    #                            prob[offset + i][0]*prob[ind, 1])
    #    term_2 = 1*prob[offset + i][1]*prob[ind, 1]
    #    print(term_1.shape, term_2.shape)
    #    return np.sum(term_0 + term_1 + term_2)

    def wprob_mixed_vectorized(self, prob, indices, distm, offset):
        """
        Compute the mixed terms

        Args:
            prob (np.array): probabilities
            indices (np.array): neighborhood structure
            distm (np.array): distance matrix
            offset (int): offsetting the indices

        Returns:
            np.array: the mixed terms
        """
        ind = indices[:, 1:]
        term_0 = 1 * prob[offset:, 0][:, None] * prob[ind][:, :, 0]
        term_1_b = (prob[offset:, 1] * prob[ind][:, :, 0].T +
                    prob[offset:, 0] * prob[ind][:, :, 1].T)
        term_1 = array_array_index(distm, ind) * term_1_b.T
        term_2 = 1 * prob[offset:, 1][:, None] * prob[ind][:, :, 1]
        return np.sum(term_0 + term_1 + term_2, axis=1)

    #def wprob_min(self, prob, i, indices, distm):
    #    term_0 = 0*prob[indices[i][1:], 0]
    #    term_1 = distm[i, indices[i][1:]]*(1*prob[indices[i][1:], 0] +
    #                                    0*prob[indices[i][1:], 1])
    #    term_2 = 1*prob[indices[i][1:], 1]
    #    return np.sum(term_0 + term_1 + term_2)

    def wprob_min_vectorized(self, prob, indices, distm):
        """
        Compute the minority terms

        Args:
            prob (np.array): probabilities
            indices (np.array): neighborhood structure
            distm (np.array): distance matrix

        Returns:
            np.array: the minority terms
        """
        ind = indices[:, 1:]
        term_0 = 0*prob[ind, 0]
        term_1 = (array_array_index(distm, ind)\
                    * (1 * prob[ind][:, :, 0] + \
                       0 * prob[ind][:, :, 1]))
        term_2 = 1*prob[ind][:, :, 1]

        return np.sum(term_0 + term_1 + term_2, axis=1)

    #def wprob_maj(self, prob, i, indices, distm):
    #    term_0 = 1*prob[indices[i][1:], 0]
    #    term_1 = distm[i, indices[i][1:]]*(0*prob[indices[i][1:], 0] +
    #                                    1*prob[indices[i][1:], 1])
    #    term_2 = 0*prob[indices[i][1:], 1]
    #    return np.sum(term_0 + term_1 + term_2)

    def wprob_maj_vectorized(self, prob, indices, distm):
        """
        Compute the majority terms

        Args:
            prob (np.array): probabilities
            indices (np.array): neighborhood structure
            distm (np.array): distance matrix

        Returns:
            np.array: the minority terms
        """
        ind = indices[:, 1:]
        term_0 = 1*prob[ind, 0]
        term_1 = array_array_index(distm, ind)\
                        *(0 * prob[ind][:, :, 0] +\
                          1 * prob[ind][:, :, 1])
        term_2 = 0*prob[ind][:, :, 1]

        return np.sum(term_0 + term_1 + term_2, axis=1)

    def utilities(self, prob, X, indices, distm):
        """
        Computes the utilit function

        Args:
            prob (np.array): strategy probabilities
            X (np.array): the training vectors
            indices (np.array): the neighborhood structure
            distm (np.array): the distance matrix

        Returns:
            np.array, np.array, np.array: utility values, minority
                                            utilities, majority
                                            utilities
        """

        #domain = range(len(X_syn))
        offset = len(X)
        #util_mixed = np.array([self.wprob_mixed(prob, i, indices, distm,
        #                                       offset=offset) for i in domain])
        util_mixed = self.wprob_mixed_vectorized(prob, indices, distm, offset)
        util_mixed = np.hstack([np.repeat(0, X.shape[0]), util_mixed])

        #util_min = np.array([self.wprob_min(prob, i, indices, distm) for i in domain])
        util_min = self.wprob_min_vectorized(prob, indices, distm)
        util_min = np.hstack([np.repeat(0, X.shape[0]), util_min])

        #util_maj = np.array([self.wprob_maj(prob, i, indices, distm) for i in domain])
        util_maj = self.wprob_maj_vectorized(prob, indices, distm)
        util_maj = np.hstack([np.repeat(0, X.shape[0]), util_maj])

        return util_mixed, util_min, util_maj

    def evolution(self,
                    *,
                    prob,
                    X,
                    X_syn, # pylint: disable=invalid-name
                    indices,
                    distm,
                    alpha=None):
        """
        Executing one step of the probabilistic evolution

        Args:
            prob (np.array): strategy probabilities
            X (np.array): original vectors
            X_syn (np.array): new samples
            indices (np.array): the neighborhood structure
            distm (np.array): the distance matrix
            alpha (float): smoothing function

        Returns:
            np.array: updated probabilities
        """
        # binary indicator indicating synthetic instances
        synthetic = np.hstack([np.array([False]*len(X)),
                                np.array([True]*len(X_syn))])

        alpha = coalesce(alpha, self.params['alpha'])

        util_mixed, util_min, util_maj = self.utilities(prob, X,
                                                        indices, distm)

        prob_new = prob.copy()
        synthetic_values = prob[:, 1] * \
            (alpha + util_min)/(alpha + util_mixed)
        prob_new[:, 1] = np.where(synthetic, synthetic_values, prob[:, 1])

        synthetic_values = prob[:, 0] * \
            (alpha + util_maj)/(alpha + util_mixed)
        prob_new[:, 0] = np.where(synthetic, synthetic_values, prob[:, 0])

        norm_factor = np.sum(prob_new, axis=1)

        prob_new[:, 0] = prob_new[:, 0]/norm_factor
        prob_new[:, 1] = prob_new[:, 1]/norm_factor

        return prob_new

    def generate_new_samples(self, X, y, nn_params):
        """
        Generate new samples.

        Args:
            X (np.array): training vectors
            y (np.array): the target labels
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array: all vectors and targets
        """
        # Applying SMOTE and ADASYN
        X_0, y_0 = SMOTE(proportion=self.proportion, # pylint: disable=invalid-name
                         n_neighbors=self.smote_n_neighbors,
                         nn_params=nn_params,
                         ss_params=self.ss_params,
                         n_jobs=self.n_jobs,
                         random_state=self._random_state_init).sample(X, y)

        X_1, y_1 = ADASYN(n_neighbors=self.params['b'], # pylint: disable=invalid-name
                          nn_params=nn_params,
                          ss_params=self.ss_params,
                          n_jobs=self.n_jobs,
                          random_state=self._random_state_init).sample(X, y)

        X_new = np.vstack([X_0, X_1[len(X):]])
        y_new = np.hstack([y_0, y_1[len(y):]])

        return X_new, y_new

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        X_all, y_all = self.generate_new_samples(X, y, nn_params)

        X_syn = X_all[len(X):] # pylint: disable=invalid-name

        # initializing strategy probabilities
        prob = np.zeros(shape=(len(X_all), 2))
        prob.fill(0.5)
        for idx in range(len(X)):
            if y[idx] == self.min_label:
                prob[idx, 0], prob[idx, 1] = 0.0, 1.0
            else:
                prob[idx, 0], prob[idx, 1] = 1.0, 0.0

        # Finding nearest neighbors, +1 as X_syn is part of X_all and nearest
        # neighbors will be themselves
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=self.params['b'] + 1,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X_all)
        indices = nnmt.kneighbors(X_syn, return_distance=False)

        # computing distances
        distm = pairwise_distances_mahalanobis(X_syn,
                                            Y=X_all,
                                            tensor=nn_params.get('metric_tensor', None))
        distm[distm == 0] = 1e-8
        distm = 1.0/distm
        distm[distm > self.params['alpha']] = self.params['alpha']

        # executing the evolution
        for _ in range(self.params['h']):
            prob = self.evolution(prob=prob,
                                    X=X,
                                    X_syn=X_syn,
                                    indices=indices,
                                    distm=distm)

        # determining final labels
        y_all[len(X):] = np.argmax(prob[len(X):], axis=1)

        return X_all, y_all

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'smote_n_neighbors': self.smote_n_neighbors,
                'b': self.params['b'],
                'alpha': self.params['alpha'],
                'h': self.params['h'],
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the NoiseFilter base class.
"""

from ..base import (StatisticsMixin, ParametersMixin)
from ..base import MetricLearningMixin

from .._logger import logger
_logger= logger

__all__= ['NoiseFilter']

class NoiseFilter(StatisticsMixin,
                  ParametersMixin,
                  MetricLearningMixin):
    """
    Parent class of noise filtering methods
    """

    def __init__(self):
        """
        Constructor
        """
        StatisticsMixin.__init__(self)
        ParametersMixin.__init__(self)
        MetricLearningMixin.__init__(self)

    def remove_noise(self, X, y):
        """
        Removes noise
        Args:
            X (np.array): features
            y (np.array): target labels
        """

    def get_params(self, deep=False):
        """
        Return parameters

        Returns:
            dict: dictionary of parameters
        """
        _ = deep
        return {}

    def set_params(self, **params):
        """
        Set parameters

        Args:
            params (dict): dictionary of parameters
        """

        for key, value in params.items():
            setattr(self, key, value)

        return self



"""
This module implements the ASMOBD method.
"""
import warnings

import numpy as np

from sklearn.preprocessing import StandardScaler
from sklearn.cluster import OPTICS

from ..base import coalesce_dict, safe_divide, fix_density, coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['ASMOBD']

class ASMOBD(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{asmobd,
                            author={Senzhang Wang and Zhoujun Li and Wenhan
                                    Chao and Qinghua Cao},
                            booktitle={The 2012 International Joint Conference
                                        on Neural Networks (IJCNN)},
                            title={Applying adaptive over-sampling technique
                                    based on data density and cost-sensitive
                                    SVM to imbalanced learning},
                            year={2012},
                            volume={},
                            number={},
                            pages={1-8},
                            doi={10.1109/IJCNN.2012.6252696},
                            ISSN={2161-4407},
                            month={June}}

    Notes:
        * In order to use absolute thresholds, the data is standardized.
        * The technique has many parameters, not easy to find the right
            combination.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_noise_removal,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 min_samples=3,
                 eps=0.8,
                 eta=0.5,
                 T_12=(1.0, 1.0),
                 t_12=(4.0, 4.0),
                 a=0.05,
                 smoothing='linear',
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            min_samples (int): parameter of OPTICS
            eps (float): parameter of OPTICS
            eta (float): tradeoff parameter
            T_12 tuple(float): noise threshold (see paper)
            t_12 tuple(float): noise threshold (see paper)
            a (float): smoothing factor (see paper)
            smoothing (str): 'sigmoid'/'linear'
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params = coalesce_dict(ss_params, {'n_dim': 2,
                                            'simplex_sampling': 'uniform',
                                            'within_simplex_sampling': 'random',
                                            'gaussian_component': None})

        super().__init__(**ss_params, random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(min_samples, "min_samples", 1)
        self.check_greater(eps, "eps", 0)
        self.check_in_range(eta, "eta", [0, 1])
        self.check_greater(a, "a", 0)
        self.check_isin(smoothing, "smoothing", ['sigmoid', 'linear'])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        # preventing too many instance attributes
        self.params = {'min_samples': min_samples,
                        'eps': eps,
                        'eta': eta,
                        'T_1': T_12[0],
                        'T_2': T_12[1],
                        't_1': t_12[0],
                        't_2': t_12[1],
                        'a': a,
                        'smoothing': smoothing}

        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'min_samples': [3],
                                  'eps': [0.3],
                                  'eta': [0.5],
                                  'T_12': [(0.7, 0.7), (0.7, 1.0), (0.7, 1.4),
                                            (1.0, 0.7), (1.0, 1.0), (1.0, 1.4),
                                            (1.4, 0.7), (1.4, 1.0), (1.4, 1.4)],
                                  't_12': [(4.0, 4.0)],
                                  'a': [0.05, 0.1],
                                  'smoothing': ['sigmoid', 'linear']}

        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def calculate_density(self, *, core_distances, ratio, n_min, n_maj, n_to_sample, params):
        """
        Calculate the density used for sampling base points.

        Args:
            core_distances (np.array): the core distances
            ratio (np.array): the n_maj/n_min ratios
            n_min (int): the n_min value
            n_maj (int): the n_maj values in the neighborhoods
            n_to_sample (int): number of samples to generate
            params (dict): the ASMOBD parameters

        Returns:
            np.array: the density for sampling
        """
        eta = params['eta']
        param_a = params['a']
        # calculating density according to the smoothing function specified
        if params['smoothing'] == 'sigmoid':
            balance_ratio = np.abs(2.0 / (1.0 + np.exp(-param_a * ratio)) - 1)
            density = eta * core_distances + (1.0 - eta)*n_maj
            density = density - balance_ratio
        else:
            density = eta * core_distances + (1.0 - eta)*n_maj
            density = eta * density + (1.0 - eta)*n_min/n_to_sample

        return density

    def determine_not_noise(self, core_distances, reachability, params):
        """
        Determine the not-noise mask

        Args:
            core_distances (np.array): the core distances
            reachability (np.array): the reachability values
            params (dict): the ASMOBD params

        Returns:
            np.array: the mask of non-noise instances
        """
        # identifying noise
        noise_1 = np.logical_and(core_distances > params['T_1'],
                                reachability > params['T_2'])

        # second constraint on noise
        mean_cd = np.mean(core_distances)
        mean_r = np.mean(reachability)
        noise_2 = np.logical_and(core_distances > mean_cd * params['t_1'],
                                reachability > mean_r * params['t_2'])

        # unifying the conditions on noise
        not_noise = np.logical_not(np.logical_or(noise_1, noise_2))

        return not_noise

    def determine_maj_min_ratios(self, *, X_min, core_distances, X_ss, y,
                                        nn_params, params):
        """
        Determine the majority/minority ratios.

        Args:
            X_min (np.array): the minority indices
            core_distances (np.array): core distances
            X_ss (np.array): the standardized base data
            y (np.array): the target labels
            nn_params (dict): the nearest neighbors parameters
            params (dict): the ASMOBD parameters

        Returns:
            np.array, np.array, np.array: the minority, majority counts
                                        and their ratios.
        """
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=params['min_samples'],
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nearestn.fit(X_ss)

        n_majs = []
        n_mins = []
        ratio = []
        for idx, row in enumerate(X_min):
            ind = nearestn.radius_neighbors(row.reshape(1, -1),
                                            radius=core_distances[idx],
                                            return_distance=False)[0]
            n_majs.append(np.sum(y[ind.astype(int)] == self.maj_label)/len(ind))
            n_mins.append(len(ind) - n_majs[-1] - 1)
            ratio.append(safe_divide(n_majs[-1], n_mins[-1], np.inf))

        n_majs = np.array(n_majs)
        n_mins = np.array(n_mins)
        ratio = np.array(ratio)

        return n_mins, n_majs, ratio

    def determine_not_noise_and_density(self, *, X_ss, y, X_min,
                                        nn_params, n_to_sample, params):
        """
        Determine the not-noise mask and the density.

        Args:
            X_ss (np.array): the standardized base data
            y (np.array): the target labels
            X_min (np.array): the minority samples
            nn_params (dict): the nearest neighbors parameters
            n_to_sample (int): the number of samples to generate
            params (dict): the ASMOBD parameters

        Returns:
            np.array, np.array: the mask of non-noise samples and the density
        """
        # executing the optics algorithm
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            optics = OPTICS(min_samples=\
                                np.min([len(X_min)-1, params['min_samples']]),
                            max_eps=params['eps'],
                            n_jobs=self.n_jobs)
            optics.fit(X_min)

        # fitting nearest neighbors models to identify the number of majority
        # samples in local environments
        n_min, n_maj, ratio = self.determine_maj_min_ratios(X_min=X_min,
                                                    core_distances=optics.core_distances_,
                                                    X_ss=X_ss,
                                                    y=y,
                                                    nn_params=nn_params,
                                                    params=params)

        not_noise = self.determine_not_noise(optics.core_distances_,
                                                optics.reachability_,
                                                params)
        density = self.calculate_density(core_distances=optics.core_distances_,
                                        ratio=ratio,
                                        n_min=n_min,
                                        n_maj=n_maj,
                                        n_to_sample=n_to_sample,
                                        params=params)

        return not_noise, density

    def determine_not_noise_and_density_safe(self, *, X_ss, y, X_min,
                                                nn_params, n_to_sample):
        """
        Determining the non-noise and density samples by releasing
        the conditions until some non-noise samples are detected.

        Args:
            X_ss (np.array): the standardized base data
            y (np.array): the target labels
            X_min (np.array): the minority samples
            nn_params (dict): the nearest neighbors parameters
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array, np.array: the mask of non-noise samples and the density
        """
        multiplier = 1.5
        params_tmp = self.params.copy()

        for _ in range(4):
            not_noise, density = self.determine_not_noise_and_density(X_ss=X_ss,
                                                        y=y,
                                                        X_min=X_min,
                                                        nn_params=nn_params,
                                                        n_to_sample=n_to_sample,
                                                        params=params_tmp)
            if np.sum(not_noise) < 2:
                params_tmp['eps'] *= multiplier
                params_tmp['T_1'] *= multiplier
                params_tmp['T_2'] *= multiplier
                params_tmp['t_1'] *= multiplier
                params_tmp['t_2'] *= multiplier
            else:
                return not_noise, density

        return not_noise, density

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """

        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        # standardizing the data to enable using absolute thresholds
        scaler = StandardScaler().fit(X)
        X_ss = scaler.transform(X) # pylint: disable=invalid-name

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X_ss, y)

        X_min = X_ss[y == self.min_label]

        not_noise, density = self.determine_not_noise_and_density_safe(X_ss=X_ss,
                                                    y=y,
                                                    X_min=X_min,
                                                    nn_params=nn_params,
                                                    n_to_sample=n_to_sample)
        # removing noise and adjusting the density factors accordingly
        X_min_not_noise = X_min[not_noise] # pylint: disable=invalid-name

        # checking if there are not-noisy samples
        if len(X_min_not_noise) <= 2:
            return self.return_copies(X, y, "not enough not-noise minority sample remained")

        density = np.delete(density, np.where(np.logical_not(not_noise))[0])
        density = fix_density(density)

        # fitting nearest neighbors model to non-noise minority samples
        n_neighbors = np.min([len(X_min_not_noise), self.params['min_samples'] + 1])
        nn_not_noise = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                        n_jobs=self.n_jobs,
                                                        **(nn_params))
        nn_not_noise.fit(X_min_not_noise)
        ind = nn_not_noise.kneighbors(X_min_not_noise, return_distance=False)

        samples = self.sample_simplex(X=X_min_not_noise,
                                        indices=ind,
                                        n_to_sample=n_to_sample,
                                        base_weights=density)

        return (np.vstack([X, scaler.inverse_transform(np.vstack(samples))]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'min_samples': self.params['min_samples'],
                'eps': self.params['eps'],
                'eta': self.params['eta'],
                'T_12': (self.params['T_1'], self.params['T_2']),
                't_12': (self.params['t_1'], self.params['t_2']),
                'a': self.params['a'],
                'smoothing': self.params['smoothing'],
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the TRIM_SMOTE method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex

from .._logger import logger
_logger= logger

__all__= ['TRIM_SMOTE']

class TRIM_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{trim_smote,
                            author="Puntumapon, Kamthorn
                            and Waiyamai, Kitsana",
                            editor="Tan, Pang-Ning
                            and Chawla, Sanjay
                            and Ho, Chin Kuan
                            and Bailey, James",
                            title="A Pruning-Based Approach for Searching
                                    Precise and Generalized Region for
                                    Synthetic Minority Over-Sampling",
                            booktitle="Advances in Knowledge Discovery
                                        and Data Mining",
                            year="2012",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="371--382",
                            isbn="978-3-642-30220-6"
                            }

    Notes:
        * It is not described precisely how the filtered data is used for
            sample generation. The method is proposed to be a preprocessing
            step, and it states that it applies sample generation to each
            group extracted.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 min_precision=0.3,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): control parameter of the nearest neighbor component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            min_precision (float): minimum value of the precision
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_in_range(min_precision, 'min_precision', [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.min_precision = min_precision
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'min_precision': [0.3]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def trim(self, y):
        """
        Determines the trim value.

        Args:
            y (np.array): array of target labels

        Returns:
            float: the trim value
        """
        return np.sum(y == self.min_label)**2/len(y)

    def precision(self, y):
        """
        Determines the precision value.

        Args:
            y (np.array): array of target labels

        Returns:
            float: the precision value
        """
        return np.sum(y == self.min_label)/len(y)

    def determine_splitting_point(self, X, y, split_on_border=False):
        """
        Determines the splitting point.

        Args:
            X (np.array): a subset of the training data
            y (np.array): an array of target labels
            split_on_border (bool): wether splitting on class borders is
                                    considered

        Returns:
            tuple(int, float), bool: (splitting feature, splitting value),
                                        make the split
        """
        trim_value = self.trim(y)
        max_t_minus_gain = 0.0
        split = None

        # checking all dimensions of X
        for idx in range(X.shape[1]):
            # sort the elements in dimension i
            sorted_X_y = sorted(zip(X[:, idx], y), key=lambda pair: pair[0]) # pylint: disable=invalid-name
            sorted_y = [yy for _, yy in sorted_X_y]

            # number of minority samples on the left
            left_min = 0
            # number of minority samples on the right
            right_min = np.sum(sorted_y == self.min_label)

            # check all possible splitting points sequentially
            for jdx in range(0, len(sorted_y)-1):
                if sorted_y[jdx] == self.min_label:
                    # adjusting the number of minority and majority samples
                    left_min = left_min + 1
                    right_min = right_min - 1
                # checking if we can split on the border and do not split
                # tie-ing feature values
                if ((split_on_border is False
                     or (split_on_border is True
                        and not sorted_y[jdx-1] == sorted_y[jdx]))
                        and sorted_X_y[jdx][0] != sorted_X_y[jdx+1][0]):
                    # compute trim value of the left
                    trim_left = left_min**2/(jdx+1)
                    # compute trim value of the right
                    trim_right = right_min**2/(len(sorted_y) - jdx - 1)
                    # let's check the gain
                    if max([trim_left, trim_right]) > max_t_minus_gain:
                        max_t_minus_gain = max([trim_left, trim_right])
                        split = (idx, sorted_X_y[jdx][0])

        # return splitting values and the value of the logical condition
        # in line 9
        if split is not None:
            return split, max_t_minus_gain > trim_value

        return (0, 0), False

    def leaf_loop(self, leafs, candidates):
        """
        The leaf loop from the paper.

        Args:
            leafs (list): the actual leafs
            candidates (list): the candidates

        Returns:
            list, list: the candidates and the potential leafs
        """
        add_to_leafs = []

        # small leafs are thrown away as noise
        leafs = [leaf for leaf in leafs if len(leaf[0]) > 1]

        # executing the loop starting in line 3
        for leaf in leafs:
            # the function implements the loop starting in line 6
            # splitting on class border is forced
            split, gain = self.determine_splitting_point(leaf[0],
                                                         leaf[1],
                                                         True)
            # condition in line 9
            if gain:
                # making the split
                mask_left = (leaf[0][:, split[0]] <= split[1])
                X_left = leaf[0][mask_left] # pylint: disable=invalid-name
                y_left = leaf[1][mask_left]
                mask_right = np.logical_not(mask_left)
                X_right = leaf[0][mask_right] # pylint: disable=invalid-name
                y_right = leaf[1][mask_right]

                # condition in line 11
                if np.sum(y_left == self.min_label) > 0:
                    add_to_leafs.append((X_left, y_left))
                # condition in line 13
                if np.sum(y_right == self.min_label) > 0:
                    add_to_leafs.append((X_right, y_right))
            else:
                # line 16
                candidates.append(leaf)

        return candidates, add_to_leafs

    def candidate_loop(self, leafs, candidates, seeds):
        """
        The candidate loop from the paper.

        Args:
            leafs (list): the actual leafs
            candidates (list): the candidates
            seeds (list): the actual seeds

        Returns:
            list, list: the updated leafs and seeds lists
        """
        # small candidates are thrown away as noise
        candidates = [cand for cand in candidates if len(cand[0]) > 1]

        # iterating through all candidates (loop starting in line 21)
        for cand in candidates:
            # extracting splitting points, this time split on border
            # is not forced
            split, gain = self.determine_splitting_point(cand[0], cand[1], False)

            # checking condition in line 27
            if gain:
                # doing the split
                mask_left = (cand[0][:, split[0]] <= split[1])
                X_left, y_left = cand[0][mask_left], cand[1][mask_left] # pylint: disable=invalid-name
                mask_right = ~mask_left
                X_right, y_right = cand[0][mask_right], cand[1][mask_right] # pylint: disable=invalid-name
                # checking logic in line 29
                if np.sum(y_left == self.min_label) > 0:
                    leafs.append((X_left, y_left))
                # checking logic in line 31
                if np.sum(y_right == self.min_label) > 0:
                    leafs.append((X_right, y_right))
            else:
                # adding candidate to seeds (line 35)
                seeds.append(cand)

        return leafs, seeds

    def trimming(self, X, y):
        """
        Do the trimming.

        Args:
            X (np.array): the training vectors
            y (np.array): the target labels

        Returns:
            list: the seeds
        """
        leafs = [(X, y)]
        candidates = []
        seeds = []

        # executing the trimming
        # loop in line 2 of the paper
        _logger.info("%s: do the trimming process",
                        self.__class__.__name__)

        while len(leafs) > 0 or len(candidates) > 0:
            candidates, add_to_leafs = self.leaf_loop(leafs, candidates)

            # we implement line 15 and 18 by replacing the list of leafs by
            # the list of new leafs.
            leafs = add_to_leafs

            leafs, seeds = self.candidate_loop(leafs, candidates, seeds)

            # line 33 and line 36 are implemented by emptying the candidates
            # list
            candidates = []

        return seeds

    def generate_samples(self, X, y,
                            X_seed_min, # pylint: disable=invalid-name
                            n_to_sample):
        """
        Generate samples.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            X_seed_min (np.array): the seed points
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        n_neighbors = min([len(X_seed_min), self.n_neighbors+1])

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_seed_min)
        indices = nnmt.kneighbors(X_seed_min, return_distance=False)

        #n_dim_orig = self.n_dim
        #self.n_dim = np.min([self.n_dim, X_seed_min.shape[0]])
        samples = self.sample_simplex(X=X_seed_min,
                                      indices=indices,
                                      n_to_sample=n_to_sample)
        #self.n_dim = n_dim_orig

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        seeds = self.trimming(X, y)

        # filtering the resulting set
        filtered_seeds = [s for s in seeds if self.precision(s[1]) > self.min_precision]

        # fix for bad choice of min_precision
        multiplier = 0.9
        while len(filtered_seeds) == 0:
            threshold = self.min_precision * multiplier
            filtered_seeds = [s for s in seeds
                                if self.precision(s[1]) > threshold]
            multiplier = multiplier * 0.9

        seeds = filtered_seeds

        X_seed = np.vstack([s[0] for s in seeds]) # pylint: disable=invalid-name
        y_seed = np.hstack([s[1] for s in seeds])

        _logger.info("%s: do the sampling", self.__class__.__name__)
        # generating samples by SMOTE
        X_seed_min = X_seed[y_seed == self.min_label] # pylint: disable=invalid-name

        samples = self.generate_samples(X, y, X_seed_min, n_to_sample)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'min_precision': self.min_precision,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}
